#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <endian.h>
#include <errno.h>
#include <inttypes.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "mu.h"


#define REQUEST_SIZE 8
#define RESPONSE_SIZE 4
#define MAX_MESSAGE_SIZE 8


static int
client_create(const char *srv_ip, const char *srv_port)
{
    struct sockaddr_in sa;
    int sk;
    int err;

    sk = socket(AF_INET, SOCK_STREAM, 0);
    if (sk == -1)
        mu_die_errno(errno, "socket");

    mu_init_sockaddr_in(&sa, srv_ip, srv_port);
    err = connect(sk, (struct sockaddr *)&sa, sizeof(sa));
    if (err == -1)
        mu_die_errno(errno, "connect");

    return sk;
}


static uint32_t
do_add_rpc(int sk, uint32_t x, uint32_t y)
{
    uint8_t msg[MAX_MESSAGE_SIZE] = { 0 };
    int err;
    size_t total = 0;
    uint32_t tmp, res;

    /* serialize request */
    x = htobe32(x);
    memcpy(msg, &x, sizeof(x));
    y = htobe32(y);
    memcpy(msg + sizeof(x), &y, sizeof(y));

    /* send request */
    err = mu_write_n(sk, msg, REQUEST_SIZE, &total);
    if (err < 0)
        mu_die_errno(-err, "send failed");

    /* receive response */
    err = mu_read_n(sk, msg, RESPONSE_SIZE, &total);
    if (err < 0)
        mu_die_errno(-err, "failed to receive response");
    else if (total != RESPONSE_SIZE)
        mu_die("server disconnected");

    /* deserialize response */
    tmp = *((uint32_t *)msg);
    res = be32toh(tmp);

    return res;
}

static void
do_random_add(unsigned int seed, const char *ip, const char *port)
{
    int sk;
    uint32_t x, y, res;

    srand(seed);

    sk = client_create(ip, port);

    x = rand() % 10;
    y = rand() % 10;

    res = do_add_rpc(sk, x, y);

    printf("client %u:  %" PRIu32 " + %" PRIu32 " = %" PRIu32 "\n",
            seed, x, y, res);

    close(sk);
}


int 
main(int argc,char *argv[])
{
    int err;
    unsigned int i =0, num_clients = 0;
    pid_t pid;

    if (argc != 4)
        mu_die("Usage: %s IP PORT NUM_CLIENTS", argv[0]);

    err = mu_str_to_uint(argv[3], 10, &num_clients);
    if (err != 0 || num_clients == 0)
        mu_die("invalid num_clients");


    for (i = 0; i < num_clients; i++) {
        pid = fork();
        if (pid == -1)
            mu_die_errno(errno, "fork");

        if (pid == 0) {
            do_random_add(i, argv[1], argv[2]);
            _exit(EXIT_SUCCESS);
        }
    }


    for (i = 0; i < num_clients; i++) {
        wait(NULL);
    }

    return 0;
}
